/*
 * Copyright (c) 2004 Damir Bikmuhametov <boco@ufanet.ru>
 * Based on kavmilter by Rashid N. Achilov <shelton@granch.ru>
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: mlfiproc.c,v 1.2 2004/02/22 09:29:31 boco Exp $
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/stat.h>

#include "common.h"
#include "mlfiproc.h"
#include "avecomm.h"
#include "kavcomm.h"

#define PRIVP ((Priv *) smfi_getpriv(ctx))

void cleanup_private(Priv *);

sfsistat mlfi_connect(SMFICTX * ctx, char * hostname, _SOCK_ADDR * hostaddr)
{
	Priv * pp = NULL;
	char * id = "mlfi_connect";

	if (!(pp = malloc(sizeof (Priv)))) {
		syslog(LOG_ERR, "%s: can't allocate private data", id);
		return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);
	} else
		memset(pp, 0, sizeof(Priv));

	smfi_setpriv(ctx, (void *) pp);

	return SMFIS_CONTINUE;
}

sfsistat mlfi_envfrom(SMFICTX * ctx, char ** argv)
{
	Priv * pp = PRIVP;
	char * id = "mlfi_envfrom";
	char * tmpname = "msg.XXXXXX";
	int tmpfd = -1;

	/* safety */

	if (!pp || !argv || !*argv || !**argv)
		return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);

	/* clear private data */

  	cleanup_private(pp);

	/* fill in "envelope from" member of private struct */

	if (!(pp->msgfrom = strdup(argv[0]))) {
		syslog(LOG_ERR, "%s: can't allocate private data", id);
		return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);
	}

	/* allocate memory for temporary file name */

	if (!(pp->fname = (char *) malloc(strlen(spooldir) + strlen(tmpname) + 2))) {
		syslog(LOG_ERR, "%s: can't allocate memory for temporary file name", id);
		cleanup_private(pp);
		return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);
	}

	/* prepare file name */

	sprintf(pp->fname, "%s/%s", spooldir, tmpname);

	/* open and chmod temporary file */

	if ((tmpfd = mkstemp(pp->fname)) == -1) {
		syslog(LOG_ERR, "%s: mkstemp error: %m, filename: %s", id, pp->fname);
		cleanup_private(pp);
		return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);
	}

	if (chmod(pp->fname, S_IWUSR | S_IRUSR | S_IRGRP ) == -1)
		syslog(LOG_ERR, "%s: chmod error: %m, filename: %s", id, pp->fname);

	/* associate stream with descriptor */

	if (!(pp->fp = fdopen(tmpfd, "w+"))) {
		syslog(LOG_ERR, "%s: fdopen error: %m, filename: %s", id, pp->fname);
		close(tmpfd);
		cleanup_private(pp);
		return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);
	}

	return SMFIS_CONTINUE;
}

sfsistat mlfi_envrcpt(SMFICTX * ctx, char ** argv)
{
	Priv * pp = PRIVP;
	char * id = "mlfi_envrcpt";

	/* safety */

	if (!pp || !argv || !*argv || !**argv)
		return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);

	if (!pp->msgto) {
		if (!(pp->msgto = strdup(argv[0]))) {
			syslog(LOG_ERR, "%s: can't allocate memory for private data", id);
			cleanup_private(pp);
			return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);
		}
	}

	return SMFIS_CONTINUE;
}

sfsistat mlfi_header(SMFICTX * ctx, char * headerf, char * headerv)
{
	Priv * pp = PRIVP;
	char * id = "mlfi_header";

	/* safety */

	if (!pp || !headerf || !headerv)
		return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);

	/* dump header to file */

	if (fprintf(pp->fp, "%s: %s\n", headerf, headerv) < 0) {
		syslog(LOG_ERR, "%s: fprintf failed: %m", id);
		cleanup_private(pp);
		return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);
	}

	return SMFIS_CONTINUE;
}

sfsistat mlfi_eoh(SMFICTX * ctx)
{
	Priv * pp = PRIVP;
	char * id = "mlfi_eoh";

	/* safety */

	if (!pp)
		return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);

	/* put delimiter */

	if (fprintf(pp->fp, "\n") < 0) {
		syslog(LOG_ERR, "%s: fprintf failed: %m", id);
		cleanup_private(pp);
		return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);
	}

	return SMFIS_CONTINUE;
}

sfsistat mlfi_body(SMFICTX * ctx, unsigned char * bodyp, size_t bodylen)
{
	Priv * pp = PRIVP;
	char * id = "mlfi_body";
	int nwritten;

	/* safety */

	if (!pp || !bodyp)
		return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);

	/* dump body part */

	nwritten = fwrite(bodyp, 1, bodylen, pp->fp);

	pp->msgsize += nwritten;

	if (nwritten != bodylen) {
		syslog(LOG_ERR, "%s: fwrite failed: %m", id);
		cleanup_private(pp);
		return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);
	}

	return SMFIS_CONTINUE;
}

sfsistat mlfi_eom(SMFICTX * ctx)
{
	Priv * pp = PRIVP;
	char * id = "mlfi_eom";
	int rc;
	char * viruslist = NULL;
	char * reply = NULL;
	const char *rejecttpl = "Message infected: ";	/* reply template */
	const char *null = "(null)";

	/* safety */

	if (!pp)
		return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);

	/* close temporary file */

	if (pp->fp) {
		fclose(pp->fp);
		pp->fp = NULL;
	}

	/* if messsage size is greater than "max message size",
	 * pass it silently */

	if (maxmsgsize && pp->msgsize > maxmsgsize * 1024) {
		syslog(LOG_INFO, "Message is too big. Skipping AV scan...");
		cleanup_private(pp);
		return SMFIS_CONTINUE;
	}

	/* check temporary file by KAV/AVE */

	rc = av_engine_protocol ? ave_scan_file(pp->fname, &viruslist) : kav_scan_file(pp->fname, &viruslist);

	if (rc == 1) {
		syslog(LOG_INFO, "DANGER! Mail from %s to %s infected: %s", pp->msgfrom, pp->msgto, (viruslist ? viruslist : null));

		/* depending on mode, reject the message or silently discard it */

		if (mode_discard) {
			if (viruslist) free(viruslist);
			cleanup_private(pp);
			return SMFIS_DISCARD;

		} else if (mode_reject) {
			if (!(reply = (char *) malloc(strlen(rejecttpl) + 1 + (viruslist ? strlen(viruslist) : strlen(null))))) {
				syslog(LOG_ERR, "%s: can't allocate memory for reply", id);
				if (viruslist) free(viruslist);
				cleanup_private(pp);
				return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);
			}
			strcpy(reply, rejecttpl);
			strcat(reply, (viruslist ? viruslist : null));
			smfi_setreply(ctx, "554", "5.7.1", reply);
			if (viruslist) free(viruslist);
			if (reply) free(reply);
			cleanup_private(pp);
			return SMFIS_REJECT;
		}
	} else if (rc == -1) {
		if (viruslist) free(viruslist);
		cleanup_private(pp);
		return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);
	}

	if (viruslist) free(viruslist);
	cleanup_private(pp);

	return SMFIS_CONTINUE;
}

sfsistat mlfi_abort(SMFICTX * ctx)
{
	Priv * pp = PRIVP;
	char * id = "mlfi_abort";

	/* safety */

	if (!pp)
		return (tempfail ? SMFIS_TEMPFAIL : SMFIS_ACCEPT);

	cleanup_private(pp);

	return SMFIS_CONTINUE;
}

sfsistat mlfi_close(SMFICTX * ctx)
{
	Priv * pp = PRIVP;
	char * id = "mlfi_close";

	/* safety */

	if (!pp)
		return SMFIS_CONTINUE;	/* strange, eh? =) */

	free(pp);
	smfi_setpriv(ctx, NULL);

	return SMFIS_CONTINUE;
}

void cleanup_private(Priv * pp)
{

	if (!pp)
		return;

	/* close temporary file */

	if (pp->fp) {
		fclose(pp->fp);
		pp->fp = NULL;
	}

	/* free allocated memory */

	if (pp->fname) {
		unlink(pp->fname);
		free(pp->fname);
		pp->fname = NULL;
	}

	if (pp->msgfrom) {
		free(pp->msgfrom);
		pp->msgfrom = NULL;
	}

	if (pp->msgto) {
		free(pp->msgto);
		pp->msgto = NULL;
	}

	pp->msgsize = 0;

	return;
}
